/*
 * grader-comm - grader-ctl.c
 *
 * Copyright (C) 2017  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <unistd.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <pu-lib.h>
#include <pu-buff.h>
#include <pu-proto.h>
#include <pu-proto-lib.h>
#include <pugcp-proto.h>
#include <pugcp-proto-lib.h>
#include <gcp.h>
#include <list.h>

#ifndef PU_MTU
#define PU_MTU		1048576
#endif

static const char *server_host = "localhost";
static const char *server_port = "10000";

static const char copyright_string[] =
"grader-ctl " PACKAGE_VERSION ".\n"
"Copyright (C) 2012  Krzysztof Mazur <krzysiek@podlesie.net>.\n";

static void show_copyright(void)
{
	puts(copyright_string);
}
	
static const char usage[] =
"usage: %s [options] [--] <id...>\n"
"Options:\n"
"      --fsync                use fsync()\n"
"  -V, --version              show version\n"
"  -h, --help                 show this help\n"
"  -p, --port=PORT            use specified port (3000)\n"
"  -s, --host=HOST            use specified target host (localhost)\n"
"\n"
"Report bugs to krzysiek@podlesie.net.\n";

static void show_usage(char *name)
{
	show_copyright();

	printf(usage, (name == NULL) ? "pu-get" : name);
	exit(EXIT_SUCCESS);
}

static const char license[] =
"This program comes with ABSOLUTELY NO WARRANTY; see COPYING for details.\n"
"This is free software, and you are welcome to redistribute it\n"
"under certain conditions; see COPYING for details.\n";

static void show_version(void)
{
	show_copyright();

	puts(license);
	exit(EXIT_SUCCESS);
}

static void parse_command_line(int argc, char **argv)
{
	static struct option long_options[] = {
		{"help", 0, NULL, 'h'},
		{"host", 1, NULL, 's'},
		{"port", 1, NULL, 'p'},
		{"version", 0, NULL, 'V'},
		{NULL, 0, NULL, 0},
	};
	int option_index = 0;
	int c;

	while (1) {
		c = getopt_long(argc, argv, "hp:s:V",
				long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case 'h':
			show_usage(argv[0]);
			BUG();
		case 'p':
			server_port = strdup(optarg);
			break;
		case 's':
			server_host = strdup(optarg);
			break;
		case 'V':
			show_version();
			BUG();
		default:
			die("unknown option: %d.", c);
		}
	}
}

static int do_connect(const char *host, const char *port)
{
	struct addrinfo hints, *ai, *aitop;
	int ret;
	int fd;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = 0;
	ret = getaddrinfo(host, port, &hints, &aitop);
	if (ret < 0)
		die("getaddrinfo(%s, %s): %s\n", host, port, gai_strerror(errno));

	for (ai = aitop; ai; ai = ai->ai_next) {
		fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
		if (fd == -1) {
			warn("socket: %s\n", strerror(errno));
			continue;
		}

		ret = connect(fd, ai->ai_addr, ai->ai_addrlen);
		if (ret < 0) {
			warn("connect: %s\n", strerror(errno));
			close(fd);
			continue;
		}
		freeaddrinfo(aitop);
		return fd;
	}
	die("can't connect to %s:%s", host, port);
}

int pu_proto_xmit(void *priv, struct pu_buff *skb)
{
	int fd = *((int *) priv);

	write_or_die(fd, skb->data, skb->len);
	pu_consume_skb(skb);
	return 0;
}

static struct pu_buff *pu_proto_recv(void *priv)
{
	struct pu_buff *skb;
	struct pu_header *hdr;
	int fd = *((int *) priv);
	unsigned long len;
	ssize_t rd;

	skb = pu_alloc_skb(512, GFP_KERNEL);
	rd = read_in_full(fd, pu_skb_put(skb, sizeof(*hdr)), sizeof(*hdr));
	if (rd != sizeof(*hdr))
		die_errno("cannot read header");

	hdr = (void *) skb->data;
	if (hdr->magic != PU_MAGIC)
		die("protocol error: invalid magic %02x", hdr->magic);

	len = be32_to_cpu(hdr->data_bytes);
	if ((len > 0) && !(len & PU_DATA_ERROR)) {
		unsigned int align = pu_align_data(len);
		struct pu_buff *skb2;

		if (align >= PU_MTU)
			die("message to large: %d bytes", len);

		skb2 = skb_copy_expand(skb, 0, align, GFP_KERNEL);
		pu_consume_skb(skb);
		skb = skb2;

		rd = read_in_full(fd, pu_skb_put(skb, align), align);
		if (rd != align)
			die_errno("cannot read payload");

		pu_skb_trim(skb, sizeof(*hdr) + len);
	}
	return skb;
}

static void *pu_proto_verify(struct pu_buff *skb, unsigned int proto,
		unsigned int id)
{
	struct pu_header *hdr = (void *) skb->data;
	unsigned int pkt_id;
	unsigned int len;

	if (hdr->proto != proto) {
		die("protocol error: invalid subprotocol: %d, expected %d",
				hdr->proto, proto);
	}

	pkt_id = be16_to_cpu(hdr->id);
	if (pkt_id != id) {
		die("protocol error: invalid id: %d, expected %d",
				pkt_id, id);
	}

	len = be32_to_cpu(hdr->data_bytes);
	if (len & PU_DATA_ERROR) {
		warn("protocol warning: id: %d, remote error: %d",
				pkt_id, len & ~PU_DATA_ERROR);
		return NULL;
	}

	return pu_skb_pull(skb, sizeof(*hdr));
}

static void *pugcp_proto_verify(struct pu_buff *skb, unsigned int proto,
		unsigned int node, unsigned int command,
		unsigned int id)
{
	struct pugcp_header *hdr = pu_proto_verify(skb, PU_PROTO_GRADER, id);
	unsigned int pkt_proto;
	unsigned int pkt_node;
	unsigned int pkt_command;

	if (!hdr)
		return NULL;

	pkt_proto = be32_to_cpu(hdr->proto);
	if (pkt_proto != proto) {
		die("protocol error: invalid GCP protocol: %d, expected %d",
				hdr->proto, proto);
	}

	pkt_node = be32_to_cpu(hdr->node);
	if (pkt_node != node) {
		die("protocol error: invalid node: %d, expected %d",
				pkt_node, node);
	}

	pkt_command = be32_to_cpu(hdr->command);
	if (pkt_command != command) {
		die("protocol error: invalid node: %d, expected %d",
				pkt_command, command);
	}
	return pu_skb_pull(skb, sizeof(*hdr));
}

static u32 grader_psched_read(int fd, unsigned int node,
		unsigned int command, unsigned int id)
{
	struct pu_buff *skb;
	struct pugcp_psched *p;
	struct pugcp_psched *resp;
	u32 ret;

	skb = pugcp_proto_alloc_skb(sizeof(*p));
	p = (void *) pu_skb_put(skb, sizeof(*p));
	if (id > GCP_PSCHED_ID_MASK)
		die("too large id '%d'\n", id);
	p->id = cpu_to_be32(id);
	p->value = cpu_to_be32(0);

	pugcp_proto_xmit_new(&fd, skb, 0, node, command, 0);

	skb = pu_proto_recv(&fd);
	resp = pugcp_proto_verify(skb, 0, node, command, 0);
	if (!resp)
		die("read error");

	ret = be32_to_cpu(resp->value);
	pu_consume_skb(skb);
	return ret;
}

static void grader_psched_write(int fd, unsigned int node,
		unsigned int command, unsigned int id, unsigned int value)
{
	struct pu_buff *skb;
	struct pugcp_psched *p;
	struct pugcp_psched *resp;

	skb = pugcp_proto_alloc_skb(sizeof(*p));
	p = (void *) pu_skb_put(skb, sizeof(*p));
	if (id > GCP_PSCHED_ID_MASK)
		die("too large id '%d'\n", id);
	p->id = cpu_to_be32(0x80000000 | id);
	p->value = cpu_to_be32(value);
	
	pugcp_proto_xmit_new(&fd, skb, 0, node, command, 0);

	skb = pu_proto_recv(&fd);
	resp = pugcp_proto_verify(skb, 0, node, command, 0);
	if (!resp)
		die("write error");
	if (be32_to_cpu(resp->id) != id)
		die("invalid returned id");
	if (be32_to_cpu(resp->value) != value)
		die("invalid returned value: written %d, read %d\n",
				value, be32_to_cpu(resp->value));
	pu_consume_skb(skb);
}

static void pu_close(int fd)
{
	int warned = 0;
	ssize_t rd;
	char c;

	shutdown(fd, SHUT_WR);
	do {
		rd = xread(fd, &c, sizeof(c));
		if (rd < 0)
			die_errno("read error");
		if (rd && !warned) {
			warn("received garbage\n");
			warned = 1;
		}
	} while (rd);
	close(fd);
}

static int parse_cmd(const char *cmd)
{
	const struct {
		const char *name;
		unsigned int id;
	} *c, cmds[] = {
		{ "sync", GCP_CPSCHED_SYNC },
		{ "start", GCP_CPSCHED_START },
		{ "mode", GCP_CPSCHED_MODE },
		{ "period", GCP_CPSCHED_PERIOD },
		{ "time-period", GCP_CPSCHED_TIME_PERIOD },
		{ "setup-time", GCP_CPSCHED_SETUP_TIME },
		{ "hold-time", GCP_CPSCHED_HOLD_TIME },
		{ "sync-time", GCP_CPSCHED_SYNC_TIME },
		{ "backoff-time", GCP_CPSCHED_BACKOFF_TIME },
		{ "start-after", GCP_CPSCHED_START_AFTER },
		{ NULL, 0 },
	};

	for (c = cmds; c->name; c++) {
		if (!strcmp(cmd, c->name))
			return c->id;
	}
	return -1;
}

int main(int argc, char **argv)
{
	int fd;

	parse_command_line(argc, argv);

	fd = do_connect(server_host, server_port);
	while (optind < argc) {
		char *cmd;
		char *value;
		int code;
		int value_code;

		cmd = strdup(argv[optind++]);
		value = strchr(cmd, '=');
		if (value) {
			char *end;

			*value = '\0';
			value++;
			value_code = strtol(value, &end, 0);
			if (*end != '\0')
				die("invalid value '%s'", value);
		}
		code = parse_cmd(cmd);
		if (code < 0)
			die("unknown option: '%s'", cmd);
		if (value) {
			grader_psched_write(fd, 14, code, 0, value_code);
		} else {
			unsigned int x;

			x = grader_psched_read(fd, 14, code, 0);
			printf("%s = %d\n", cmd, x);
		}
		free(cmd);
	}
	pu_close(fd);
	return 0;
}
